package net.sourceforge.vrapper.vim.modes;

import static net.sourceforge.vrapper.keymap.vim.ConstructorWrappers.key;
import net.sourceforge.vrapper.keymap.KeyStroke;
import net.sourceforge.vrapper.keymap.SpecialKey;
import net.sourceforge.vrapper.platform.CommandLineUI;
import net.sourceforge.vrapper.platform.SearchAndReplaceService;
import net.sourceforge.vrapper.platform.TextContent;
import net.sourceforge.vrapper.utils.Position;
import net.sourceforge.vrapper.utils.Search;
import net.sourceforge.vrapper.utils.SearchOffset;
import net.sourceforge.vrapper.utils.SearchResult;
import net.sourceforge.vrapper.utils.SubstitutionDefinition;
import net.sourceforge.vrapper.vim.EditorAdaptor;
import net.sourceforge.vrapper.vim.Options;
import net.sourceforge.vrapper.vim.commands.CenterLineCommand;
import net.sourceforge.vrapper.vim.commands.CommandExecutionException;
import net.sourceforge.vrapper.vim.commands.motions.StickyColumnPolicy;

/**
 * Prompt the user for each match in a substitution, activated by the 'c' flag
 * of a substitution definition (e.g., :s/foo/bar/c).
 */
public class ConfirmSubstitutionMode extends AbstractMode {

    public static final String NAME = "confirm substitution mode";
    public static final String DISPLAY_NAME = "CONFIRM SUBSTITUTION";
    private static final KeyStroke KEY_ESCAPE = key(SpecialKey.ESC);
    
    private SubstitutionDefinition subDef;
    private int endOffset;
    private int nextLine;
    private int totalLength;
    private int replaceOffset = 0;
    private boolean globalFlag = false;
    private boolean caseSensitive = false;
    private boolean useRegex = true;
    private CommandLineUI commandLine;
    private SearchResult lastMatch;

    public ConfirmSubstitutionMode(EditorAdaptor editorAdaptor) {
        super(editorAdaptor);
    }

    @Override
    public String getName() {
        return NAME;
    }

    @Override
    public String getDisplayName() {
        return DISPLAY_NAME;
    }
    
    /**
     * ModeSwitchHint used when entering mode
     * REQUIRED for ConfirmSubstitutionMode
     */
    public static class SubstitutionConfirm implements ModeSwitchHint {
        protected SubstitutionDefinition subDef;
        protected int startLine;
        protected int endLine;

        public SubstitutionConfirm(SubstitutionDefinition subDef, int startLine, int endLine) {
            this.subDef = subDef;
            this.startLine = startLine;
            this.endLine = endLine;
        }
    }

    @Override
    public void enterMode(ModeSwitchHint... hints) throws CommandExecutionException {
        super.enterMode();
        if (hints.length > 0 && hints[0] instanceof SubstitutionConfirm) {
            SubstitutionConfirm hint = (SubstitutionConfirm) hints[0];
            subDef = hint.subDef;
            globalFlag = subDef.flags.contains("g");
            caseSensitive = editorAdaptor.getSearchAndReplaceService().isCaseSensitive(subDef.find, subDef.flags);
            useRegex = editorAdaptor.getConfiguration().get(Options.SEARCH_REGEX);
            nextLine = hint.startLine;

            TextContent model = editorAdaptor.getModelContent();
            totalLength = model.getTextLength();
            if(hint.endLine == model.getNumberOfLines()) {
                endOffset = totalLength;
            }
            else {
                endOffset = model.getLineInformation(hint.endLine).getEndOffset();
            }
        }
        else {
            //a SubstitutionConfirm instance is required to enter this mode
            exit();
            return;
        }
        
        commandLine = editorAdaptor.getCommandLine();
        commandLine.setPrompt("replace with " + subDef.replace + " (y/n/a/q/l)?");
        commandLine.open();
        
        findNextMatch(true);
    }
    
    @Override
    public void leaveMode(ModeSwitchHint... hints) throws CommandExecutionException {
        super.leaveMode(hints);
        commandLine.close();
    }
    
    private void findNextMatch(boolean doHighlight) {
        int startOffset;
        if(lastMatch != null && globalFlag) {
            //next match might be on the same line
            startOffset = lastMatch.getRightBound().getModelOffset() + replaceOffset;
            //performing a substitution could change our next starting position
            //(so we don't match on the replaced string) reset replaceOffset in
            //case the user hits 'n' and we *don't* modify starting position
            replaceOffset = 0;
        }
        else {
            //start on next line
            startOffset = editorAdaptor.getModelContent().getLineInformation(nextLine).getBeginOffset();
        }
    
        SearchOffset afterSearch = new SearchOffset.Begin(startOffset);
        Search search = new Search(subDef.find, false, caseSensitive, afterSearch, useRegex);
        SearchAndReplaceService searchService = editorAdaptor.getSearchAndReplaceService();
        Position start = editorAdaptor.getCursorService().newPositionForModelOffset(startOffset);
        SearchResult result = searchService.find(search, start);
        //if no match found or match is outside our range
        if(result.getStart() == null || result.getRightBound().getModelOffset() > endOffset) {
            exit();
            return;
        }

        if (result.getLeftBound().getViewOffset() < 0) {
            editorAdaptor.getViewportService().exposeModelPosition(result.getLeftBound());
        }
        
        editorAdaptor.setPosition(result.getLeftBound(), StickyColumnPolicy.NEVER);
        if(doHighlight) {
            //force match to be visible (move scrollbars)
            //is there a better way to do this?
            CenterLineCommand.CENTER.execute(editorAdaptor);
            //highlight match
            searchService.incSearchhighlight(result.getLeftBound(), result.getModelLength());
        }

        //prepare for next iteration
        lastMatch = result;
        nextLine = editorAdaptor.getModelContent().getLineInformationOfOffset(
                lastMatch.getRightBound().getModelOffset()
                ).getNumber() + 1; //next line after match
    }

    @Override
    public boolean handleKey(KeyStroke stroke) {
        if(stroke.equals(KEY_ESCAPE)) {
            exit();
            return true;
        }

        switch(stroke.getCharacter()) {
        case 'y':
            performSubstitution();
            findNextMatch(true);
            break;
        case 'n':
            findNextMatch(true);
            break;
        case 'a':
            replaceAll();
            exit();
            break;
        case 'l':
            performSubstitution();
            exit();
            break;
        case 'q':
            exit();
            break;
        default:
            //ignore all other key presses
            break;
        }
        return true;
    }
    
    private void exit() {
        lastMatch = null;
        editorAdaptor.getSearchAndReplaceService().removeIncSearchHighlighting();
        editorAdaptor.changeModeSafely(NormalMode.NAME);
    }
    
    private void replaceAll() {
        editorAdaptor.getHistory().beginCompoundChange();
        while(lastMatch != null) {
            performSubstitution();
            findNextMatch(false);
        }
        editorAdaptor.getHistory().endCompoundChange();
    }

    private void performSubstitution() {
        editorAdaptor.getSearchAndReplaceService().substitute(
                lastMatch.getLeftBound().getModelOffset(),
                subDef.find, subDef.flags, subDef.replace);
        
        int newTotal = editorAdaptor.getModelContent().getTextLength();
        //keep offsets in sync when find and replace are different lengths
        replaceOffset = newTotal - totalLength;
        endOffset += replaceOffset;
        totalLength = newTotal;
    }

    @Override
    public String resolveKeyMap(KeyStroke stroke) {
        return null;
    }

}
